from __future__ import annotations

from typing import Any

from plain import models

from . import validators
from .hashers import (
    hash_password,
    identify_hasher,
)


class PasswordField(models.CharField):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs["max_length"] = 128
        kwargs.setdefault(
            "validators",
            [
                validators.MinimumLengthValidator(),
                validators.CommonPasswordValidator(),
                validators.NumericPasswordValidator(),
            ],
        )
        super().__init__(*args, **kwargs)

    def deconstruct(self) -> tuple[str | None, str, list[Any], dict[str, Any]]:
        name, path, args, kwargs = super().deconstruct()
        if kwargs.get("max_length") == 128:
            del kwargs["max_length"]
        return name, path, args, kwargs

    def pre_save(self, model_instance: models.Model, add: bool) -> str | None:
        value = super().pre_save(model_instance, add)

        if value and not self._is_hashed(value):
            value = hash_password(value)
            # Set the hashed value back on the instance immediately too
            setattr(model_instance, self.attname, value)

        return value

    @staticmethod
    def _is_hashed(value: str) -> bool:
        try:
            identify_hasher(value)
            return True
        except ValueError:
            return False



from __future__ import annotations

from collections.abc import Callable, Generator
from typing import TYPE_CHECKING, Any

from plain import forms
from plain.auth import get_user_model
from plain.exceptions import ValidationError
from plain.models.forms import ModelForm

from .core import check_user_password
from .hashers import check_password
from .utils import unicode_ci_compare

if TYPE_CHECKING:
    from plain.models import Model


class PasswordResetForm(forms.Form):
    email = forms.EmailField(max_length=254)

    def send_mail(
        self,
        *,
        template_name: str,
        context: dict[str, Any],
        from_email: str,
        to_email: str,
    ) -> None:
        from plain.email import TemplateEmail

        email = TemplateEmail(
            template=template_name,
            context=context,
            from_email=from_email,
            to=[to_email],
            headers={
                "X-Auto-Response-Suppress": "All",
            },
        )

        email.send()

    def get_users(self, email: str) -> Generator[Model, None, None]:
        """Given an email, return matching user(s) who should receive a reset.

        This allows subclasses to more easily customize the default policies
        that prevent inactive users and users with unusable passwords from
        resetting their password.
        """
        active_users = get_user_model().query.filter(email__iexact=email)
        return (u for u in active_users if unicode_ci_compare(email, u.email))

    def save(
        self,
        *,
        generate_reset_url: Callable[[Model], str],
        email_template_name: str = "password_reset",
        from_email: str = "",
        extra_email_context: dict[str, Any] | None = None,
    ) -> None:
        """
        Generate a one-use only link for resetting password and send it to the
        user.
        """
        email = self.cleaned_data["email"]
        for user in self.get_users(email):
            context = {
                "email": email,
                "user": user,
                "url": generate_reset_url(user),
                **(extra_email_context or {}),
            }
            self.send_mail(
                template_name=email_template_name,
                context=context,
                from_email=from_email,
                to_email=user.email,  # type: ignore[attr-defined]
            )


class PasswordSetForm(forms.Form):
    """
    A form that lets a user set their password without entering the old
    password
    """

    new_password1 = forms.CharField(strip=False)
    new_password2 = forms.CharField(strip=False)

    def __init__(self, user: Model, *args: Any, **kwargs: Any) -> None:
        self.user = user
        super().__init__(*args, **kwargs)

    def clean_new_password2(self) -> str:
        password1 = self.cleaned_data.get("new_password1")
        password2 = self.cleaned_data.get("new_password2")
        if password1 and password2 and password1 != password2:
            raise ValidationError(
                "The two password fields didn't match.",
                code="password_mismatch",
            )

        # password2 must exist at this point (required field)
        assert isinstance(password2, str), "new_password2 must be a string"

        # Clean it as if it were being put into the model directly
        self.user._model_meta.get_field("password").clean(password2, self.user)  # type: ignore[union-attr]

        return password2

    def save(self, commit: bool = True) -> Model:
        # User model has password attribute
        self.user.password = self.cleaned_data["new_password1"]  # type: ignore[assignment]
        if commit:
            self.user.save()
        return self.user


class PasswordChangeForm(PasswordSetForm):
    """
    A form that lets a user change their password by entering their old
    password.
    """

    current_password = forms.CharField(strip=False)

    def clean_current_password(self) -> str:
        """
        Validate that the current_password field is correct.
        """
        current_password = self.cleaned_data["current_password"]
        if not check_user_password(self.user, current_password):
            raise ValidationError(
                "Your old password was entered incorrectly. Please enter it again.",
                code="password_incorrect",
            )
        return current_password


class PasswordLoginForm(forms.Form):
    email = forms.EmailField(max_length=150)
    password = forms.CharField(strip=False)

    def clean(self) -> dict[str, Any]:
        User = get_user_model()

        email = self.cleaned_data.get("email")
        password = self.cleaned_data.get("password")

        if email and password:
            try:
                # The vast majority of users won't have a case-sensitive email, so we act that way
                user = User.query.get(email__iexact=email)
            except User.DoesNotExist:
                # Run the default password hasher once to reduce the timing
                # difference between an existing and a nonexistent user (django #20760).
                check_password(password, "")

                raise ValidationError(
                    "Please enter a correct email and password. Note that both fields may be case-sensitive.",
                    code="invalid_login",
                )

            if not check_user_password(user, password):
                raise ValidationError(
                    "Please enter a correct email and password. Note that both fields may be case-sensitive.",
                    code="invalid_login",
                )

            self._user = user

        return self.cleaned_data

    def get_user(self) -> Model:
        return self._user


class PasswordSignupForm(ModelForm):
    confirm_password = forms.CharField(strip=False)

    class Meta:
        model = get_user_model()
        fields = ("email", "password")

    def clean(self) -> dict[str, Any]:
        cleaned_data = super().clean()
        password = cleaned_data.get("password")
        confirm_password = cleaned_data.get("confirm_password")
        if password and confirm_password and password != confirm_password:
            raise ValidationError(
                "The two password fields didn't match.",
                code="password_mismatch",
            )
        return cleaned_data



from __future__ import annotations

import base64
import unicodedata
from binascii import Error as BinasciiError


def urlsafe_base64_encode(s: bytes) -> str:
    """
    Encode a bytestring to a base64 string for use in URLs. Strip any trailing
    equal signs.
    """
    return base64.urlsafe_b64encode(s).rstrip(b"\n=").decode("ascii")


def urlsafe_base64_decode(s: str) -> bytes:
    """
    Decode a base64 encoded string. Add back any trailing equal signs that
    might have been stripped.
    """
    s_bytes = s.encode()
    try:
        return base64.urlsafe_b64decode(
            s_bytes.ljust(len(s_bytes) + len(s_bytes) % 4, b"=")
        )
    except (LookupError, BinasciiError) as e:
        raise ValueError(e)


def unicode_ci_compare(s1: str, s2: str) -> bool:
    """
    Perform case-insensitive comparison of two identifiers, using the
    recommended algorithm from Unicode Technical Report 36, section
    2.11.2(B)(2).
    """
    return (
        unicodedata.normalize("NFKC", s1).casefold()
        == unicodedata.normalize("NFKC", s2).casefold()
    )



from __future__ import annotations

import hmac
from datetime import datetime
from typing import TYPE_CHECKING, Any

from plain import signing
from plain.auth import get_user_model
from plain.auth.sessions import login as auth_login
from plain.auth.sessions import update_session_auth_hash
from plain.auth.views import AuthView
from plain.forms import BaseForm
from plain.http import (
    BadRequestError400,
    RedirectResponse,
)
from plain.urls import reverse
from plain.utils.cache import add_never_cache_headers
from plain.utils.encoding import force_bytes
from plain.views import CreateView, FormView

from .forms import (
    PasswordChangeForm,
    PasswordLoginForm,
    PasswordResetForm,
    PasswordSetForm,
    PasswordSignupForm,
)

if TYPE_CHECKING:
    from plain.http import Response
    from plain.models import Model


class PasswordForgotView(FormView):
    form_class = PasswordResetForm
    reset_confirm_url_name: str

    def generate_password_reset_token(self, user: Any) -> str:
        return signing.dumps(
            {
                "id": user.id,
                "email": user.email,
                "password": user.password,  # Hashed password
                "timestamp": datetime.now().timestamp(),  # Makes each token unique
            },
            salt="password-reset",
            compress=True,
        )

    def generate_password_reset_url(self, user: Any) -> str:
        token = self.generate_password_reset_token(user)
        url = reverse(self.reset_confirm_url_name) + f"?token={token}"
        return self.request.build_absolute_uri(url)

    def form_valid(self, form: PasswordResetForm) -> Response:  # type: ignore[override]
        form.save(
            generate_reset_url=self.generate_password_reset_url,
        )
        return super().form_valid(form)


class PasswordResetView(AuthView, FormView):
    form_class = PasswordSetForm
    reset_token_max_age = 60 * 60  # 1 hour
    _reset_token_session_key = "_password_reset_token"

    def check_password_reset_token(self, token: str) -> Model | None:
        max_age = self.reset_token_max_age

        try:
            data = signing.loads(token, salt="password-reset", max_age=max_age)
        except signing.SignatureExpired:
            return None
        except signing.BadSignature:
            return None

        UserModel = get_user_model()
        try:
            user = UserModel.query.get(id=data["id"])
        except (TypeError, ValueError, OverflowError, UserModel.DoesNotExist):
            return None

        # If the password has changed since the token was generated, the token is invalid.
        # (These are the hashed passwords, not the raw passwords.)
        if not hmac.compare_digest(
            force_bytes(user.password), force_bytes(data["password"])
        ):
            return None

        # If the email has changed since the token was generated, the token is invalid.
        if not hmac.compare_digest(force_bytes(user.email), force_bytes(data["email"])):
            return None

        return user

    def get(self) -> Response:
        if self.user:
            # Redirect if the user is already logged in
            return RedirectResponse(str(self.success_url) if self.success_url else "/")

        # Tokens are initially passed as GET parameters and we
        # immediately store them in the session and remove it from the URL.
        if token := self.request.query_params.get("token", ""):
            # Store the token in the session and redirect to the
            # password reset form at a URL without the token. That
            # avoids the possibility of leaking the token in the
            # HTTP Referer header.
            self.session[self._reset_token_session_key] = token
            # Redirect to the path itself, without the GET parameters
            response = RedirectResponse(self.request.path)
            add_never_cache_headers(response)
            return response

        return super().get()

    def get_user(self) -> Model:
        session_token = self.session.get(self._reset_token_session_key, "")
        if not session_token:
            # No token in the session, so we can't check the password reset token.
            raise BadRequestError400("No password reset token found.")

        user = self.check_password_reset_token(session_token)
        if not user:
            # Remove it from the session if it is invalid.
            del self.session[self._reset_token_session_key]
            raise BadRequestError400("Password reset token is no longer valid.")

        return user

    def get_form_kwargs(self) -> dict:
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.get_user()
        return kwargs

    def form_valid(self, form: PasswordSetForm) -> Response:  # type: ignore[override]
        form.save()
        del self.session[self._reset_token_session_key]
        # If you wanted, you could log in the user here so they don't have to
        # go through the log in form again.
        return super().form_valid(form)


class PasswordChangeView(AuthView, FormView):
    # Change to PasswordSetForm if you want to set new passwords
    # without confirming the old one.
    form_class = PasswordChangeForm

    def get_form_kwargs(self) -> dict:
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.user
        return kwargs

    def form_valid(self, form: PasswordChangeForm) -> Response:  # type: ignore[override]
        form.save()
        # Updating the password logs out all other sessions for the user
        # except the current one.
        update_session_auth_hash(self.request, form.user)
        return super().form_valid(form)


class PasswordLoginView(AuthView, FormView):
    form_class = PasswordLoginForm
    success_url = "/"

    def get(self) -> Response:
        # Redirect if the user is already logged in
        if self.user:
            return RedirectResponse(self.success_url)

        return super().get()

    def form_valid(self, form: PasswordLoginForm) -> Response:  # type: ignore[override]
        # Log the user in and redirect
        auth_login(self.request, form.get_user())

        return super().form_valid(form)


class PasswordSignupView(CreateView):
    form_class = PasswordSignupForm
    success_url = "/"

    def form_valid(self, form: BaseForm) -> Response:
        # # Log the user in and redirect
        # auth_login(self.request, form.save())

        return super().form_valid(form)



from __future__ import annotations

from typing import Any

from .hashers import check_password, hash_password


def check_user_password(user: Any, password: str) -> bool:
    # Run the default password hasher once to reduce the timing
    # difference between an existing and a nonexistent user (#20760).
    hash_password(password)

    # Update the stored hashed password if the hashing algorithm changed
    def setter(raw_password: str) -> None:
        user.password = raw_password
        user.save(update_fields=["password"])

    password_is_correct = check_password(password, user.password, setter)

    return password_is_correct



from __future__ import annotations

import base64
import functools
import hashlib
import hmac
import math
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any

from plain.exceptions import ImproperlyConfigured
from plain.runtime import settings
from plain.utils.crypto import (
    RANDOM_STRING_CHARS,
    get_random_string,
    pbkdf2,
)
from plain.utils.encoding import force_bytes
from plain.utils.module_loading import import_string


def check_password(
    password: str,
    encoded: str,
    setter: Callable[[str], None] | None = None,
    preferred: str | BasePasswordHasher = "default",
) -> bool:
    """
    Return a boolean of whether the raw password matches the three
    part encoded digest.

    If setter is specified, it'll be called when you need to
    regenerate the password.
    """
    if not password:
        return False

    preferred = get_hasher(preferred)
    try:
        hasher = identify_hasher(encoded)
    except ValueError:
        # encoded is gibberish or uses a hasher that's no longer installed.
        return False

    hasher_changed = hasher.algorithm != preferred.algorithm
    must_update = hasher_changed or preferred.must_update(encoded)
    is_correct = hasher.verify(password, encoded)

    # If the hasher didn't change (we don't protect against enumeration if it
    # does) and the password should get updated, try to close the timing gap
    # between the work factor of the current encoded password and the default
    # work factor.
    if not is_correct and not hasher_changed and must_update:
        hasher.harden_runtime(password, encoded)

    if setter and is_correct and must_update:
        setter(password)
    return is_correct


def hash_password(
    password: str,
    salt: str | None = None,
    hasher: str | BasePasswordHasher = "default",
) -> str:
    """
    Turn a plain-text password into a hash for database storage

    Same as encode() but generate a new random salt. If password is None then
    return a concatenation of UNUSABLE_PASSWORD_PREFIX and a random string,
    which disallows logins. Additional random string reduces chances of gaining
    access to admin or superuser accounts. See ticket #20079 for more info.
    """
    hasher = get_hasher(hasher)
    salt = salt or hasher.salt()
    return hasher.encode(password, salt)


@functools.lru_cache
def get_hashers() -> list[BasePasswordHasher]:
    hashers = []
    for hasher_path in settings.PASSWORD_HASHERS:
        hasher_cls = import_string(hasher_path)
        hasher = hasher_cls()
        if not getattr(hasher, "algorithm"):
            raise ImproperlyConfigured(
                f"hasher doesn't specify an algorithm name: {hasher_path}"
            )
        hashers.append(hasher)
    return hashers


@functools.lru_cache
def get_hashers_by_algorithm() -> dict[str, BasePasswordHasher]:
    return {
        hasher.algorithm: hasher
        for hasher in get_hashers()
        if hasher.algorithm is not None
    }


def get_hasher(algorithm: str | BasePasswordHasher = "default") -> BasePasswordHasher:
    """
    Return an instance of a loaded password hasher.

    If algorithm is 'default', return the default hasher. Lazily import hashers
    specified in the project's settings file if needed.
    """
    if isinstance(algorithm, BasePasswordHasher):
        return algorithm

    elif algorithm == "default":
        return get_hashers()[0]

    else:
        hashers = get_hashers_by_algorithm()
        try:
            return hashers[algorithm]
        except KeyError:
            raise ValueError(
                f"Unknown password hashing algorithm '{algorithm}'. "
                "Did you specify it in the PASSWORD_HASHERS "
                "setting?"
            )


def identify_hasher(encoded: str) -> BasePasswordHasher:
    """
    Return an instance of a loaded password hasher.

    Identify hasher algorithm by examining encoded hash, and call
    get_hasher() to return hasher. Raise ValueError if
    algorithm cannot be identified, or if hasher is not loaded.
    """
    # Ancient versions of Plain created plain MD5 passwords and accepted
    # MD5 passwords with an empty salt.
    if (len(encoded) == 32 and "$" not in encoded) or (
        len(encoded) == 37 and encoded.startswith("md5$$")
    ):
        algorithm = "unsalted_md5"
    # Ancient versions of Plain accepted SHA1 passwords with an empty salt.
    elif len(encoded) == 46 and encoded.startswith("sha1$$"):
        algorithm = "unsalted_sha1"
    else:
        algorithm = encoded.split("$", 1)[0]
    return get_hasher(algorithm)


def mask_hash(hash: str, show: int = 6, char: str = "*") -> str:
    """
    Return the given hash, with only the first ``show`` number shown. The
    rest are masked with ``char`` for security reasons.
    """
    masked = hash[:show]
    masked += char * len(hash[show:])
    return masked


def must_update_salt(salt: str, expected_entropy: int) -> bool:
    # Each character in the salt provides log_2(len(alphabet)) bits of entropy.
    return len(salt) * math.log2(len(RANDOM_STRING_CHARS)) < expected_entropy


class BasePasswordHasher(ABC):
    """
    Abstract base class for password hashers

    When creating your own hasher, you need to override algorithm,
    verify(), encode() and safe_summary().

    PasswordHasher objects are immutable.
    """

    algorithm: str | None = None
    salt_entropy: int = 128

    def salt(self) -> str:
        """
        Generate a cryptographically secure nonce salt in ASCII with an entropy
        of at least `salt_entropy` bits.
        """
        # Each character in the salt provides
        # log_2(len(alphabet)) bits of entropy.
        char_count = math.ceil(self.salt_entropy / math.log2(len(RANDOM_STRING_CHARS)))
        return get_random_string(char_count, allowed_chars=RANDOM_STRING_CHARS)

    @abstractmethod
    def verify(self, password: str, encoded: str) -> bool:
        """Check if the given password is correct."""
        ...

    def _check_encode_args(self, password: str, salt: str) -> None:
        if password is None:
            raise TypeError("password must be provided.")
        if not salt or "$" in salt:
            raise ValueError("salt must be provided and cannot contain $.")

    @abstractmethod
    def encode(self, password: str, salt: str) -> str:
        """
        Create an encoded database value.

        The result is normally formatted as "algorithm$salt$hash" and
        must be fewer than 128 characters.
        """
        ...

    @abstractmethod
    def decode(self, encoded: str) -> dict[str, Any]:
        """
        Return a decoded database value.

        The result is a dictionary and should contain `algorithm`, `hash`, and
        `salt`. Extra keys can be algorithm specific like `iterations` or
        `work_factor`.
        """
        ...

    @abstractmethod
    def safe_summary(self, encoded: str) -> dict[str, Any]:
        """
        Return a summary of safe values.

        The result is a dictionary and will be used where the password field
        must be displayed to construct a safe representation of the password.
        """
        ...

    def must_update(self, encoded: str) -> bool:
        return False

    def harden_runtime(self, password: str, encoded: str) -> None:
        """
        Bridge the runtime gap between the work factor supplied in `encoded`
        and the work factor suggested by this hasher.

        Taking PBKDF2 as an example, if `encoded` contains 20000 iterations and
        `self.iterations` is 30000, this method should run password through
        another 10000 iterations of PBKDF2. Similar approaches should exist
        for any hasher that has a work factor. If not, this method should be
        defined as a no-op to silence the warning.
        """
        warnings.warn(
            "subclasses of BasePasswordHasher should provide a harden_runtime() method"
        )


class PBKDF2PasswordHasher(BasePasswordHasher):
    """
    Secure password hashing using the PBKDF2 algorithm (recommended)

    Configured to use PBKDF2 + HMAC + SHA256.
    The result is a 64 byte binary string.  Iterations may be changed
    safely but you must rename the algorithm if you change SHA256.
    """

    algorithm = "pbkdf2_sha256"
    iterations = 720000
    digest = hashlib.sha256

    def encode(self, password: str, salt: str, iterations: int | None = None) -> str:
        self._check_encode_args(password, salt)
        iterations = iterations or self.iterations
        hash = pbkdf2(password, salt, iterations, digest=self.digest)
        hash = base64.b64encode(hash).decode("ascii").strip()
        return "%s$%d$%s$%s" % (self.algorithm, iterations, salt, hash)  # noqa: UP031

    def decode(self, encoded: str) -> dict[str, Any]:
        algorithm, iterations, salt, hash = encoded.split("$", 3)
        assert algorithm == self.algorithm
        return {
            "algorithm": algorithm,
            "hash": hash,
            "iterations": int(iterations),
            "salt": salt,
        }

    def verify(self, password: str, encoded: str) -> bool:
        decoded = self.decode(encoded)
        encoded_2 = self.encode(password, decoded["salt"], decoded["iterations"])
        return hmac.compare_digest(force_bytes(encoded), force_bytes(encoded_2))

    def safe_summary(self, encoded: str) -> dict[str, Any]:
        decoded = self.decode(encoded)
        return {
            "algorithm": decoded["algorithm"],
            "iterations": decoded["iterations"],
            "salt": mask_hash(decoded["salt"]),
            "hash": mask_hash(decoded["hash"]),
        }

    def must_update(self, encoded: str) -> bool:
        decoded = self.decode(encoded)
        update_salt = must_update_salt(decoded["salt"], self.salt_entropy)
        return (decoded["iterations"] != self.iterations) or update_salt

    def harden_runtime(self, password: str, encoded: str) -> None:
        decoded = self.decode(encoded)
        extra_iterations = self.iterations - decoded["iterations"]
        if extra_iterations > 0:
            self.encode(password, decoded["salt"], extra_iterations)
